@@ -5,7 +5,7 @@ import torch.nn.functional as F
 from deprecated import deprecated
 
 from torch_gauge.o3 import O3Tensor, SphericalTensor
-from torch_gauge.o3.functional import NormContraction1d
+from torch_gauge.o3.functional import NormContraction1d, SumsqrContraction1d
 
 
 class SSP(torch.nn.Softplus):
@@ -276,6 +276,186 @@ class IELin(torch.nn.Module):
         )
 
 
+class BlockSparseIELin(torch.nn.Module):
+    r"""
+    Sparisfied irrep-wise Equivariant Linear Layer. Recommended for building large scale models.
+
+    Linear transformation within each irrep are approximated via a random permutation on the m indices
+    and learnable block-diagonal linear layers
+
+    The permutation and matrix multiplication are performed on the last dimension of the spherical tensors.
+
+    Attributes:
+        metadata_in (torch.LongTensor): the number of non-degenerate feature channels of the input tensor
+            for each l (and the associated m(s)).
+        metadata_out (torch.LongTensor): the number of non-degenerate feature channels of the output tensor
+            for each l (and the associated m(s)).
+        group (str): The group index of the tensors to be passed.
+        block_size (int): size of diagonal blocks in linear layers, must be a common denominator of metadata_out
+    """
+
+    def __init__(self, metadata_in, metadata_out, group="o3", block_size=1):
+        super().__init__()
+        assert metadata_in.dim() == 1
+        assert len(metadata_in) == len(metadata_out)
+        self._metadata_in = metadata_in
+        self._metadata_out = metadata_out
+        group = group.lower()
+        if group == "so3":
+            self.tensor_class = SphericalTensor
+            self.n_irreps_per_l = torch.arange(start=0, end=metadata_in.size(0)) * 2 + 1
+        elif group == "o3":
+            self.tensor_class = O3Tensor
+            n_irreps_per_l = torch.arange(start=0, end=metadata_in.size(0) // 2) * 2 + 1
+            self.n_irreps_per_l = n_irreps_per_l.repeat_interleave(2)
+        else:
+            raise NotImplementedError(f"The group {group} is not supported in IELin")
+
+        self.block_size = block_size
+        self._metadata_inter = metadata_out.clone()
+        self._metadata_inter[(self._metadata_in == 0)] = 0
+        self.register_buffer(
+            "interim_layout",
+            self.tensor_class.generate_rep_layout_1d_(self._metadata_inter),
+        )
+        self.interim_size = self.interim_layout.shape[1]
+        assert torch.remainder(self._metadata_inter, self.block_size).sum() == 0, print(
+            "The block_size much be a gcd of irrep sizes, got remainders: ",
+            torch.remainder(self._metadata_inter, self.block_size),
+        )
+        self._isometric = False
+        if torch.all(self._metadata_in == self._metadata_out):
+            self._isometric = True
+        if not self._isometric:
+            self.scaling_factors = torch.nn.Parameter(
+                torch.rand(self._metadata_inter.sum().item()) + 0.5
+            )
+
+        # random permutation or scatter_add + small linear kernel
+        irrep_scatter_idx = []
+        vecin_select_idx = []
+        vecout_select_mask = []
+        src_offset = 0
+        interim_offset = 0
+        for lp_idx, nin_lpm in enumerate(metadata_in):
+            nin_lpm = nin_lpm.item()
+            nout_lpm = metadata_out[lp_idx].item()
+            if nout_lpm == 0:
+                # Pointer shift on input tensor
+                src_offset += nin_lpm * self.n_irreps_per_l[lp_idx]
+                continue
+            if nin_lpm == 0:
+                # Pointer shift on output tensor
+                vecout_select_mask.append(
+                    torch.zeros(
+                        nout_lpm * self.n_irreps_per_l[lp_idx], dtype=torch.bool
+                    )
+                )
+                continue
+            zoom_factor = -(nin_lpm // (-nout_lpm))
+            segment_scatter_idx = torch.arange(nout_lpm, dtype=torch.long).repeat(
+                zoom_factor
+            )
+            segment_scatter_idx = segment_scatter_idx[
+                torch.randperm(nout_lpm * zoom_factor)
+            ][:nin_lpm]
+            vecout_select_mask.append(
+                torch.ones(nout_lpm * self.n_irreps_per_l[lp_idx], dtype=torch.bool)
+            )
+            for m in range(self.n_irreps_per_l[lp_idx]):
+                vecin_select_idx_lpm = (
+                    torch.arange(nin_lpm, dtype=torch.long) + src_offset
+                )
+                vecin_select_idx.append(vecin_select_idx_lpm)
+                irrep_scatter_idx.append(segment_scatter_idx + interim_offset)
+                src_offset += nin_lpm
+                interim_offset += nout_lpm
+        self.register_buffer("vecin_select_idx", torch.cat(vecin_select_idx))
+        self.register_buffer("irrep_scatter_idx", torch.cat(irrep_scatter_idx))
+        self.register_buffer("vecout_select_mask", torch.cat(vecout_select_mask))
+
+        if self.block_size > 1:
+            self.linear = torch.nn.Linear(self.block_size, self.block_size, bias=False)
+
+        self.register_buffer(
+            "out_layout",
+            self.tensor_class.generate_rep_layout_1d_(self._metadata_out),
+        )
+        assert self.interim_size == interim_offset
+        assert self.out_layout.shape[1] == self.vecout_select_mask.shape[0]
+        self.num_out_channels = torch.sum(self._metadata_out).item()
+        # self.reset_parameters()
+
+    def reset_parameters(self) -> None:
+        with torch.no_grad():
+            bound = math.sqrt(1 / self.block_size)
+            self.linear.weight.data.uniform_(-bound, bound)
+
+    def forward(self, x: SphericalTensor) -> SphericalTensor:
+        """
+        Args:
+            x: The input SphericalTensor.
+
+        Returns:
+            The data tensor of the output spherical tensor.
+        """
+        assert len(x.rep_dims) == 1
+        assert x.rep_dims[-1] == x.ten.dim() - 1
+        assert torch.all(x.metadata[-1].eq(self._metadata_in)), (
+            f"Expected the SphericalTensor x and self._metadata_in to have the "
+            f"same irrep metadata along the last dimension, got {x.metadata[-1]}"
+            f" and {self._metadata_in} instead"
+        )
+        in_ten = x.ten.view(-1, x.ten.shape[-1])
+        if self._isometric:
+            interim_ten = torch.index_select(in_ten, 1, self.irrep_scatter_idx)
+            if self.block_size > 1:
+                interim_ten = (
+                    self.linear(interim_ten.view(in_ten.shape[0], -1, self.block_size))
+                ).view(in_ten.shape[0], -1)
+            out_ten = interim_ten.view(*x.ten.shape[:-1], -1)
+        else:
+            selected_in_ten = torch.index_select(
+                in_ten, dim=1, index=self.vecin_select_idx
+            )
+            interim_ten = (
+                torch.zeros(
+                    in_ten.shape[0],
+                    self.interim_size,
+                    dtype=in_ten.dtype,
+                    device=in_ten.device,
+                )
+                .index_add_(1, self.irrep_scatter_idx, selected_in_ten)
+                .mul(self.scaling_factors[self.interim_layout[0]].unsqueeze(0))
+            )
+            if self.block_size > 1:
+                interim_ten = (
+                    self.linear(interim_ten.view(in_ten.shape[0], -1, self.block_size))
+                ).view(in_ten.shape[0], -1)
+            out_ten = (
+                torch.zeros(
+                    in_ten.shape[0],
+                    self.out_layout.shape[1],
+                    dtype=in_ten.dtype,
+                    device=in_ten.device,
+                )
+                .masked_scatter(
+                    self.vecout_select_mask.unsqueeze(0),
+                    interim_ten,
+                )
+                .view(*x.ten.shape[:-1], -1)
+            )
+
+        out_metadata = self._metadata_out.unsqueeze(0)
+        return self.tensor_class(
+            out_ten,
+            rep_dims=x.rep_dims,
+            metadata=out_metadata,
+            rep_layout=x.rep_layout[:-1] + (self.out_layout.data,),
+            num_channels=x.num_channels[:-1] + (self.num_out_channels,),
+        )
+
+
 class IELinSerial(torch.nn.Module):
     r"""
     Irrep-wise Equivariant Linear Layer.
@@ -397,9 +577,119 @@ class IELinSerial(torch.nn.Module):
         )
 
 
+class EvNorm1d(torch.nn.Module):
+    r"""
+    Equivariant Normalization layer.
+
+    The specified invariant channels are not affected by squashing.
+    """
+
+    def __init__(
+        self,
+        num_channels,
+        n_invariant_channels=0,
+    ):
+        super().__init__()
+        self._num_channels = num_channels
+        self._n_invariant_channels = n_invariant_channels
+
+    def forward(self, x: SphericalTensor) -> (torch.Tensor, SphericalTensor):
+        """
+        Args:
+            x: The input SphericalTensor.
+
+        Returns:
+            (tuple): tuple containing:
+                (torch.Tensor): The normalized invariant content.
+
+                (torch.Tensor): The "pure gauge" tensor for equivariant channels.
+        """
+        if self._n_invariant_channels == 0:
+            divisor = (
+                SumsqrContraction1d.apply(
+                    x.ten,
+                    x.rep_layout[0][2],
+                    (x.ten.shape[0], x.num_channels[0]),
+                    1,
+                )
+                .add(1)
+                .sqrt()
+            )
+            x1 = divisor - 1
+            divisor_broadcasted = torch.index_select(
+                divisor,
+                dim=1,
+                index=x.rep_layout[0][2],
+            )
+            x2 = x.ten.div(divisor_broadcasted)
+        else:
+            assert self._n_invariant_channels <= x.metadata[0][0]
+            x10 = x.ten[:, : self._n_invariant_channels]
+            divisor = (
+                SumsqrContraction1d.apply(
+                    x.ten[:, self._n_invariant_channels :],
+                    x.rep_layout[0][2, self._n_invariant_channels :]
+                    - self._n_invariant_channels,
+                    (x.ten.shape[0], x.num_channels[0] - self._n_invariant_channels),
+                    1,
+                )
+                .add(1)
+                .sqrt()
+            )
+            x11 = divisor - 1
+            x1 = torch.cat([x10, x11], dim=1)
+            divisor_broadcasted = torch.index_select(
+                divisor,
+                dim=1,
+                index=(
+                    x.rep_layout[0][2, self._n_invariant_channels :]
+                    - self._n_invariant_channels
+                ),
+            )
+            x2 = x.ten[:, self._n_invariant_channels :].div(divisor_broadcasted)
+        return x1, x2
+
+
+class EvMLP(torch.nn.Module):
+    def __init__(
+        self,
+        metadata,
+        norm1,
+        norm2,
+        activation_func,
+    ):
+        super().__init__()
+        self.n_invariant_channels = metadata[0].item()
+        self.num_channels = torch.sum(metadata).item()
+        self.evnorm = EvNorm1d(self.num_channels, self.n_invariant_channels)
+        if norm2 is not None:
+            self.mlp = torch.nn.Sequential(
+                torch.nn.Linear(
+                    self.num_channels, self.n_invariant_channels, bias=False
+                ),
+                norm1,
+                torch.nn.Linear(self.n_invariant_channels, self.n_invariant_channels),
+                activation_func,
+                norm2,
+                torch.nn.Linear(self.n_invariant_channels, self.n_invariant_channels),
+            )
+        else:
+            self.mlp = torch.nn.Sequential(
+                torch.nn.Linear(
+                    self.num_channels, self.n_invariant_channels, bias=False
+                ),
+                norm1,
+            )
+
+    def forward(self, x: SphericalTensor) -> SphericalTensor:
+        x1, x2 = self.evnorm(x)
+        x1_updated = self.mlp(x1)
+        return x.self_like(torch.cat([x1_updated, x2], dim=-1))
+
+
 class RepNorm1d(torch.nn.Module):
     r"""
-    The (experimental) Representation Normalization layer.
+    The Representation Normalization layer.
 
     .. math::
 
@@ -421,12 +711,14 @@ class RepNorm1d(torch.nn.Module):
         eps=1e-2,
         n_invariant_channels=0,
         mode="raw",
+        invariant_mode="l2",
     ):
         super().__init__()
         self._num_channels = num_channels
         self._n_invariant_channels = n_invariant_channels
         self._eps = eps
         self._mode = mode
+        self._invariant_mode = invariant_mode
         if norm == "batch":
             self.norm = torch.nn.BatchNorm1d(
                 num_features=num_channels, momentum=momentum, affine=False
@@ -445,7 +737,7 @@ class RepNorm1d(torch.nn.Module):
             )
         elif self._mode == "inv":
             self.betainv = torch.nn.Parameter(
-                0.5 + torch.rand(self._num_channels - self._n_invariant_channels) * 1.5
+                torch.rand(self._num_channels - self._n_invariant_channels) * 10.0
             )
         else:
             raise NotImplementedError
@@ -462,9 +754,20 @@ class RepNorm1d(torch.nn.Module):
                 (SphericalTensor): The "pure gauge" spherical tensor.
         """
         if self.norm is None:
-            x0 = NormContraction1d.apply(
-                x.ten, x.rep_layout[0][2], (x.ten.shape[0], x.num_channels[0]), 1, 1e-04
-            )
+            if self._invariant_mode == "l2":
+                x0 = NormContraction1d.apply(
+                    x.ten,
+                    x.rep_layout[0][2],
+                    (x.ten.shape[0], x.num_channels[0]),
+                    1,
+                    1e-04,
+                )
+            elif self._invariant_mode == "sumsqr":
+                x0 = SumsqrContraction1d.apply(
+                    x.ten, x.rep_layout[0][2], (x.ten.shape[0], x.num_channels[0]), 1
+                )
+            else:
+                raise NotImplementedError
             return x0, x
         assert x.ten.dim() == 2
         if self._mode == "raw":
@@ -474,9 +777,20 @@ class RepNorm1d(torch.nn.Module):
         else:
             raise NotImplementedError
         if self._n_invariant_channels == 0:
-            x0 = NormContraction1d.apply(
-                x.ten, x.rep_layout[0][2], (x.ten.shape[0], x.num_channels[0]), 1, 1e-04
-            )
+            if self._invariant_mode == "l2":
+                x0 = NormContraction1d.apply(
+                    x.ten,
+                    x.rep_layout[0][2],
+                    (x.ten.shape[0], x.num_channels[0]),
+                    1,
+                    1e-04,
+                )
+            elif self._invariant_mode == "sumsqr":
+                x0 = SumsqrContraction1d.apply(
+                    x.ten, x.rep_layout[0][2], (x.ten.shape[0], x.num_channels[0]), 1
+                )
+            else:
+                raise NotImplementedError
             x1 = self.norm(x0)
             divisor = x0.add(beta).abs().add(self._eps)
             divisor_broadcasted = torch.index_select(
@@ -486,12 +800,18 @@ class RepNorm1d(torch.nn.Module):
             )
             x2_ten = x.ten.div(divisor_broadcasted)
         else:
-            x0 = x.invariant()
+            x0 = x.invariant(mode=self._invariant_mode)
             assert self._n_invariant_channels <= x.metadata[0][0]
             x1 = self.norm(x0)
-            divisor = (
-                torch.abs(x0[:, self._n_invariant_channels :].add(beta)) + self._eps
-            )
+            if self._invariant_mode == "uest":
+                divisor = (
+                    x0[:, self._n_invariant_channels :].pow(2).add(1 + beta.pow(2))
+                    + self._eps ** 2
+                ).sqrt()
+            else:
+                divisor = (
+                    torch.abs(x0[:, self._n_invariant_channels :].add(beta)) + self._eps
+                )
             divisor_broadcasted = torch.index_select(
                 divisor,
                 dim=1,
